#!/bin/env python
"""Generate sut java implementation from uppaal model"""
# Author: Harco Kuppens


import sys, os,  pprint, re, getpass, tempfile, subprocess, shutil  # standard libraries
import argparse # only in standard lib for python 2.7 and later -> installed local copy  (pure python versions)
import jinja2, yaml  # external libraries   -> installed local copy   (pure python versions) (not yaml doesn't use libyaml but pure python version)
import pyuppaal  # local libraries          -> installed local copy   (pure python versions)



import model_parser
import get_models

import gentemplate





#-------------------------------------------------------------------------------
#   settings
#-------------------------------------------------------------------------------
def settings(args):
    global CONFIG
    CONFIG={}

    # configure :
    # - default parameters used in learning
    #   they can be overruled by settings in the config
    CONFIG['DEFAULT_LEARNING_PARAMS']={
      # learner
      'min_value':0,
      'max_value':255,
      'seed':1299777356020l,
      'maxNumTraces': 3000,
    	'minTraceLength':20,
    	'maxTraceLength':100,
      # lotos
      'lts_min_value':0,
      'lts_max_value':255,
      'lts_nat_type':'uint8_t',
      'lts_int_type':'int8_t'
    }
    # - input models/ directory
    script_dir=os.path.abspath(os.path.dirname(args[0]))

    CONFIG['ROOT_DIR'] =  os.path.abspath( os.path.join(script_dir,'../..') )


    CONFIG['MODELS_DIR']=os.path.abspath( script_dir + '/../../models/' )
    # - directory where all output results are stored
    CONFIG['RESULTS_DIR']=os.path.abspath( script_dir + "/../../results" )

    CONFIG['RESOURCES_DIR']=os.path.abspath( script_dir + "/../../share/resources" )

    # - output sut file
    CONFIG['OUTPUT_SUT_FILE']= os.path.abspath("/tmp/Sut.java")
    # - output sut file
    CONFIG['OUTPUT_CONFIG_FILE']=  os.path.abspath("/tmp/config.yaml")

    import platform
    current_os = platform.system()
    #note: current_os.startswith('CYGWIN') is only true if using cygwin's python
    #      and not when using standard window's python from cygwin bash shell
    if current_os == 'Linux' or current_os == 'Darwin' or  current_os.startswith('CYGWIN'):
        CONFIG['java']='java'
        CONFIG['javac']='javac'
    else:
        CONFIG['java']='java.exe'
        CONFIG['javac']='javac.exe'

    check_program(CONFIG['java'])
    check_program(CONFIG['javac'])

    # Note:
    # config DEFAULT_LEARNING_PARAMS are specified with only default value
    # without the need to specify the values type explicitly because
    # we only use the python basic types of the values : int,string and boolean
    # Meaning :
    # - the code using config params knows these types!!
    # - config value in uppaal declaration is converted to basic python
    #   of the default config value




#-------------------------------------------------------------------------------
#    module  functions
#-------------------------------------------------------------------------------




#-------------------------------------------------------------------------------
#   helpers
#-------------------------------------------------------------------------------


def check_program(program):
    if not which(program):
      print 'Error 1: cannot find' +  program + 'in PATH'
      exit(1)

def which(program):
    import os

    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def print_error(return_code,command):
       error_msg = "problem running command : \n   "  + join_command(command)
       print "\n\nError " + str(return_code) + " : " + error_msg
       exit(return_code)


def join_command(command):
    strings=[]
    for str in command:
        if ' ' in str:    # put " qoutes arround string
          str=str.replace('"', '\\"')
          strings.append('"'+str+'"')
        else:
          strings.append(str)
    return ' '.join(strings)

def run_command(command,cwd=None):
    try:
      process = subprocess.Popen(command,cwd=cwd)
    except Exception as inst:
      return_code=1
      print_error(return_code,command)

    process.communicate()  # no pipes set for stdin/stdout/stdout streams so does effectively only just wait for process ends  (same as process.wait()

    if process.returncode:
      print_error(process.returncode,command)

def run_command_with_output(command,input=None,cwd=None):
    try:
      process = subprocess.Popen(command,cwd=cwd,stdout=subprocess.PIPE,stdin=subprocess.PIPE)
    except Exception as inst:
      return_code=1
      print_error(return_code,command)

    [stdoutdata, stderrdata]=process.communicate(input)  # no pipes set for stdin/stdout/stdout streams so does effectively only just wait for process ends  (same as process.wait()

    if process.returncode:
      print stderrdata
      print_error(process.returncode,command)

    return stdoutdata

def check_return_code(return_code,error_msg):
    if return_code:
       print "\n\nError " + str(return_code) + " : " + error_msg
       exit(return_code)

# def check_command_available(command):
#     args=['which',' "print \'hallo\'" ']
#     process = subprocess.Popen(["which",args])
#     return_code=process.wait()
#     check_return_code(return_code,"problem running print")
#

# create dir if not yes exists
def create_dir(path_name):
    if not os.path.isdir(path_name):
        os.mkdir(path_name)

#-------------------------------------------------------------------------------
#   steps
#-------------------------------------------------------------------------------


# copy jars
# generate sut.java + config.yaml
# javac sut.java
def generate_learner(model_file,model_results_dir,template_dir):

    outdir=os.path.join(model_results_dir, 'learner')
    srcdir=os.path.join(CONFIG['RESOURCES_DIR'] , 'learner')
    if os.path.isdir(outdir): shutil.rmtree(outdir)
    shutil.copytree(srcdir,outdir)

    import uppaal2sut

    sut_file= os.path.join(outdir , 'sut/Sut.java')
    config_file = os.path.join(outdir , 'config.yaml')

    default_config_params=CONFIG['DEFAULT_LEARNING_PARAMS']

    uppaal2sut.uppaal2sut(model_file,template_dir,sut_file,config_file,default_config_params)

    run_command([CONFIG['javac'], '-cp', 'sut.jar' , 'sut/Sut.java'],cwd=outdir)



def run_learner(model_results_dir):

    cwd=os.path.join(model_results_dir, 'learner')

    javapath='.' + os.pathsep + 'sut.jar'
    sut=[CONFIG['java'], '-cp' , javapath , 'sut.Main']
    learner=[CONFIG['java'] , '-jar' , 'learner.jar']

    try:
      command=sut
      process1 = subprocess.Popen(command,cwd=cwd)
    except Exception as inst:
      return_code=1
      print_error(return_code,command)


    try:
      command=learner
      process2 = subprocess.Popen(command,cwd=cwd)
    except Exception as inst:
      return_code=1
      print_error(return_code,command)

    process1.wait()
    process2.wait()

def learneroutput2uppaal(model):
    pass

def uppaal2lotos(model_file,destdir,template_data,template_dir):

    # cp lotos stuff in destdir
    srcdir=os.path.join(CONFIG['RESOURCES_DIR'] , 'lotos')
    if os.path.isdir(destdir): shutil.rmtree(destdir)
    shutil.copytree(srcdir,destdir)

    # generate lotos model
    command=[CONFIG['java'], '-jar', os.path.join( CONFIG['ROOT_DIR'] , 'share/java/Uppaal2Lotos.jar'), model_file]
    output=run_command_with_output(command,input=None,cwd=destdir)
    lotos_file=os.path.join( destdir , 'model.lotos')
    f=open(lotos_file,'w')
    f.write(output)
    f.close()

    # add settings for lotos in model.t file
    out_file=os.path.join( destdir , 'model.t')
    gentemplate.fill_template(out_file,template_data,template_dir,'model.t')



#-------------------------------------------------------------------------------
#   test
#-------------------------------------------------------------------------------

def test_run_command():
    #program = "ppython"
    program = "python"

    python_lines=[
      "import time",
      "print 'start'",
      "time.sleep(2)",
      'print "end"'
    ]

    args=["-c" , ';'.join(python_lines) ]
    command=[ program ] + args

    #args=['python','-c','import time;print "start";time.sleep(2);print "end";']
    print "run command in child"
    run_command( command )
    print "back at parent"



#-------------------------------------------------------------------------------
#   main + input parameters parsing
#-------------------------------------------------------------------------------


def args_parsing(args,models_dir):
    global CONFIG

    models=get_models.get_models(models_dir)

    #parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]),epilog=get_models_help(models))  # prints epilog always on single line
    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))

    argparser.add_argument('model', help='see below for available models')
    str=argparser.format_help()

    def myhelp( ):
        print str  + "\n" +  get_models.get_models_help(models)

    def myerror( message):
            print >>sys.stderr, ('%s: error: %s\n') % (argparser.prog, message)
            argparser.print_usage(sys.stderr)
            print "\n" +  get_models.get_models_help(models)
            exit(1)

    argparser.error=myerror
    argparser.format_help=myhelp

    parse_result=argparser.parse_args(args[1:])

    model_file=parse_result.model

    # if model_file is not a file path to a xml file, we assume
    # only the model's name is give as input!
    if not model_file.endswith('.xml'):
       model_name=model_file
       model_file= models_dir  + '/' + model_name + ".xml"
       model_file=os.path.abspath(model_file)

    model_file.replace('\\','/')


    return  model_file


def test_run_command_with_output():
    command=['/bin/ls','/home/']
    print run_command_with_output(command,input=None,cwd=None)


def main(args):

    sys.argv=args   # to be sure
    settings(args)

    models_dir=CONFIG['MODELS_DIR']
    sut_file = CONFIG['OUTPUT_SUT_FILE']
    config_file= CONFIG['OUTPUT_CONFIG_FILE']
    default_config_params= CONFIG['DEFAULT_LEARNING_PARAMS']

    model_file=args_parsing(args,models_dir)
    # strip basename and extensions of model_file : model_name
    model_name=os.path.basename(model_file).split('.')[0]

    script_dir=os.path.abspath(os.path.dirname(args[0]))

    template_dir=script_dir+'/templates/'




    # create results dir for current model
    create_dir(CONFIG['RESULTS_DIR'])
    model_results_dir=os.path.join(CONFIG['RESULTS_DIR'], model_name)
    create_dir(model_results_dir)

    generate_learner(model_file,model_results_dir,template_dir)

    run_learner(model_results_dir)


    # convert learner output to uppaal model
    import learnlib2uppaal
    learner_dir=os.path.join(model_results_dir, 'learner')
    json_filename_in=os.path.join(learner_dir, 'output.json')
    dot_filename_in=os.path.join(learner_dir, 'learnresult.dot')
    yaml_filename_in=os.path.join(learner_dir, 'config.yaml')

    learned_model = os.path.join(model_results_dir, 'learned.xml')
    teacher_model = os.path.join(model_results_dir, 'teacher.xml')

    shutil.copy(yaml_filename_in, os.path.join(model_results_dir, 'config.yaml')  )
    shutil.copy(model_file, teacher_model  )

    learned_model_str=learnlib2uppaal.learnlib2uppaal(json_filename_in,dot_filename_in,yaml_filename_in)
    f=open(learned_model,'w')
    f.write(learned_model_str)
    f.close()

    #os.path.join(model_results_dir, 'teacher.xml')
    #learneroutput2uppaal(model_name)



    # convert uppaal model 2 lotos



    lotos_dir=os.path.join(model_results_dir, 'lotos')
    # create lotos dir if not yet existing
    create_dir(lotos_dir)
    shutil.copy( os.path.join(CONFIG['RESOURCES_DIR'],'compare_lotos_models.sh') , os.path.join(lotos_dir,'compare_lotos_models.sh')  )


    configParams=yaml.load(file(yaml_filename_in))['configParams']
    uppaal2lotos(teacher_model,os.path.join(lotos_dir,'teacher'),configParams,template_dir)
    uppaal2lotos(learned_model,os.path.join(lotos_dir,'learned'),configParams,template_dir)




    print "done"



if __name__ == "__main__":
    main(sys.argv)


