#!/usr/bin/env python
"""Generate sut java implementation from uppaal model"""
# Author: Harco Kuppens


import sys, os,  pprint, re, getpass, tempfile   # standard libraries
import argparse # only in standard lib for python 2.7 and later -> installed local copy  (pure python versions)
import yaml  # external libraries   -> installed local copy   (pure python versions) (not yaml doesn't use libyaml but pure python version)
import pyuppaal  # local libraries          -> installed local copy   (pure python versions)
#import pygraphviz


import model_parser
import get_models
import gentemplate

#-------------------------------------------------------------------------------
#   settings for uppaal2sut
#-------------------------------------------------------------------------------
def settings(args):
    global CONFIG
    CONFIG={}

    # configure :
    # - default parameters used in learning
    #   they can be overruled by settings in the config
    CONFIG['DEFAULT_LEARNING_PARAMS']={
      # learner
      'min_value':0,
      'max_value':255,
      'seed':1299777356020l,
      'maxNumTraces': 3000,
    	'minTraceLength':20,
    	'maxTraceLength':100,
      # lotos
      'lts_min_value':0,
      'lts_max_value':255,
      'lts_nat_type':'uint8_t',
      'lts_int_type':'int8_t'
    }
    # - input models/ directory
    script_dir=os.path.abspath(os.path.dirname(args[0]))
    CONFIG['MODELS_DIR']=os.path.abspath( script_dir + '/../../models/' )
    # - output sut file
    CONFIG['OUTPUT_SUT_FILE']= os.path.abspath("/tmp/Sut.java")
    # - output sut file
    CONFIG['OUTPUT_CONFIG_FILE']=  os.path.abspath("/tmp/config.yaml")

    # Note:
    # config DEFAULT_LEARNING_PARAMS are specified with only default value
    # without the need to specify the values type explicitly because
    # we only use the python basic types of the values : int,string and boolean
    # Meaning :
    # - the code using config params knows these types!!
    # - config value in uppaal declaration is converted to basic python
    #   of the default config value


#-------------------------------------------------------------------------------
#   uppaal2sut module  functions
#-------------------------------------------------------------------------------


def write_config(output_file,data):
    selected_keys=('inputInterfaces','outputInterfaces','constants','configParams')
    #dictionary comprehension only supported for python 2.7 and higher
    #config={k:v for k,v in data.iteritems() if k in selected_keys}
    config={}
    for k,v in data.iteritems():
        if k in selected_keys:
            config[k]=v

    with open(output_file,'w') as f:
      f.write(yaml.safe_dump(config,default_flow_style=False))
      #f.write(yaml.safe_dump(config))





def uppaal2sut(model_file,template_dir,sut_file,config_file,default_config_params):
    parsed_data=model_parser.parse_uppaal_model(model_file,default_config_params)
    parsed_data['model_file']=model_file.replace('\\','\\\\')

    write_config(config_file,parsed_data)
    print "config written: " + config_file

    gentemplate.fill_template(sut_file,parsed_data,template_dir,'Sut.java')
    print "sut written: " + sut_file



#-------------------------------------------------------------------------------
#   main + input parameters parsing
#-------------------------------------------------------------------------------


def args_parsing(args,models_dir):
    global CONFIG

    models=get_models.get_models(models_dir)

    #parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]),epilog=get_models_help(models))  # prints epilog always on single line
    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))

    argparser.add_argument('model', help='see below for available models')
    str=argparser.format_help()

    def myhelp( ):
        print str  + "\n" +  get_models.get_models_help(models)

    def myerror( message):
            print >>sys.stderr, ('%s: error: %s\n') % (argparser.prog, message)
            argparser.print_usage(sys.stderr)
            print "\n" +  get_models.get_models_help(models)
            exit(1)

    argparser.error=myerror
    argparser.format_help=myhelp

    parse_result=argparser.parse_args(args[1:])

    model_file=parse_result.model

    # if model_file is not a file path to a xml file, we assume
    # only the model's name is give as input!
    if not model_file.endswith('.xml'):
       model_name=model_file
       model_file= models_dir  + '/' + model_name + ".xml"
       model_file=os.path.abspath(model_file)

    model_file.replace('\\','/')


    return  model_file


def main(args):
    sys.argv=args   # to be sure
    settings(args)

    models_dir=CONFIG['MODELS_DIR']
    sut_file = CONFIG['OUTPUT_SUT_FILE']
    config_file= CONFIG['OUTPUT_CONFIG_FILE']
    default_config_params= CONFIG['DEFAULT_LEARNING_PARAMS']

    model_file=args_parsing(args,models_dir)
    script_dir=os.path.abspath(os.path.dirname(args[0]))
    template_dir=script_dir+'/templates/'

    uppaal2sut(model_file,template_dir,sut_file,config_file,default_config_params)


if __name__ == "__main__":
    main(sys.argv)



